# Multiton module that ensures only one object to be allocated for a given
# argument list.
#
# The 'multiton' pattern is similar to a singleton, but instead of only one
# instance, there are several similar instances.  it's usefull when you want to
# avoid constructing objects many times because of some huge expence (connecting
# to a database for example), require a set of similar but not identical
# objects, and cannot easily control how many time a contructor may be called.
#
# Usage:
# class SomeMultitonClass
#   include Multiton
#   attr :arg
#   def initialize(arg)
#     @arg = arg
#   end
# end
#
# a = SomeMultitonClass.instance(4)
# b = SomeMultitonClass.instance(4)	# a and b are same object
# c = SomeMultitonClass.instance(2)     # c is a different object
#
# print([a.arg,b.arg].max, c.arg, "\n")
#
# a = SomeMultitonClass.new		# error (`new' is private)

module Multiton
  def Multiton.append_features(klass)
    #klass.private_class_method(:new)
    klass.instance_eval <<-CODE 
      alias __new__ new 
      @__instances__ = {} 
      def instance(*args, &block)
	__instance__ = @__instances__[args]
	unless __instance__
	  Thread.critical = true
	  begin
	    __instance__ = (@__instances__[args] = __new__(*args, &block))
	  ensure
	    Thread.critical = false
	  end
	end
	return __instance__
      end
      alias new instance
    CODE
  end
end

if __FILE__ == $0
  class SomeMultitonClass
    include Multiton
    attr :arg
    def initialize(arg)
      @arg = arg
    end
  end

  class File 
    include Multiton
  end

  a = SomeMultitonClass.instance(4)
  b = SomeMultitonClass.new(4)	        # a and b are same object
  c = SomeMultitonClass.instance(2)     # c is a different object

  print([a.arg,b.arg].max, c.arg, "\n") # -> 42

  File.open(__FILE__) do |f|
    lines = f.readlines
    p lines[74 - 1] # -> "  # line 74\n"
    p lines[75 - 1] # -> "  # line 75\n"
  end 
  # line 74
  # line 75

  a = File.new(__FILE__)
  b = File.new(__FILE__)
  p a == b # -> true
  
  73.times{a.gets}

  p a.gets # -> "  # line 74\n"
  p b.gets # -> "  # line 75\n"
end
